from carrot_mc.colors import Colorizer as clr

class CliEventPrinter:
    def handle(self, event: str, payload=None):
        method_name = 'handle_' + event.replace(' ', '_')
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            method(payload)

    def handle_info(self, payload):
        print(payload)

    def handle_info_mod_resolved(self, payload):
        print(f'{clr.mod_key(payload.mod_key)} resolved as {clr.mod_name(payload.mod.name)}')

    def handle_info_all_mod_check_complete(self, payload):
        print('Mod check phase complete. Proceeding to download...')

    def handle_info_will_upgrade_mod(self, payload):
        print('Mod already installed, found new version and will upgrade because it\'s allowed.')

    def handle_info_will_downgrade_mod(self, payload):
        print('Mod already installed, found older version and will downgrade because it\'s allowed.')

    def handle_info_dependencies_detected(self, payload):
        print('Detected dependencies:', payload.deps)

    def handle_info_downloading_file(self, payload):
        print(f'Downloading file {clr.file_name(payload.file.file_name)} from {clr.url(payload.file.download_url)}...')

    def handle_info_all_mod_fetch_complete(self, payload):
        print('Download phase complete. Proceeding to installation...')

    def handle_info_installing_mod(self, payload):
        if payload.new_mod:
            print(f'Installing mod {clr.mod_name(payload.mod.name)} with file {clr.file_name(payload.mod.file.file_name)}...')
        else:
            print(f'Updating mod {clr.mod_name(payload.mod.name)}...')

    def handle_info_updating_file(self, payload):
        if payload.enabled:
            print(f'Installing new file {clr.file_name(payload.file.file_name)}...')
        else:
            print(f'Installing new file {clr.file_name(payload.file.file_name + ".disabled")} because current file was also {clr.file_name(".disabled")}...')

    def handle_info_all_mod_install_complete(self, payload):
        print('Installation phase complete.')

    def handle_match_list(self, payload):
        print(f'No mod found in top downloaded mods matching exactly the key "{clr.mod_key(payload.mod_key)}". These are the top downloaded matches:')
        for mod in payload.mods:
            print(f'[{clr.mod_key(mod.key)}] {clr.mod_name(mod.name)} by {clr.mod_owner(mod.owner)}')
            print(f'\t{clr.mod_blurb(mod.blurb)}')
            if mod.categories:
                print('\t' + ', '.join([f'{clr.mod_category(c)}' for c in mod.categories]))

    def handle_mod_enabled(self, payload):
        print(f'Mod {clr.mod_name(payload.mod.name)} {clr.mod_key("[" + payload.mod.key + "]")} enabled.')

    def handle_mod_disabled(self, payload):
        print(f'Mod {clr.mod_name(payload.mod.name)} {clr.mod_key("[" + payload.mod.key + "]")} disabled.')

    def handle_warn_no_files_in_channel(self, payload):
        print('Mod has no files in the chosen channel. Skipping.')

    def handle_warn_upgrade_not_allowed(self, payload):
        print(f'A newer file was found but upgrades are disabled by default. Use the {clr.cli("--upgrade")} option if this should be allowed.')
        if payload.dependency:
            print(f'\n{clr.emph("NOTE")}: Because this is a dependency, it will {clr.emph("not")} be re-checked if you re-run last install command. Use "{clr.cli("carrot update " + payload.mod.key)}" to do it explicitly.')

    def handle_warn_downgrade_not_allowed(self, payload):
        print(f'An older file was found but downgrades are disabled by default. Use the {clr.cli("--downgrade")} option if this was intended.')

    def handle_warn_already_newest_version(self, payload):
        print('Mod already at newest version for selected channel.')

    def handle_warn_mod_already_enabled(self, payload):
        print(f'Mod {clr.mod_name(payload.mod.name)} {clr.mod_key("[" + payload.mod.key + "]")} is already enabled.')

    def handle_warn_mod_already_disabled(self, payload):
        print(f'Mod {clr.mod_name(payload.mod.name)} {clr.mod_key("[" + payload.mod.key + "]")} is already disabled.')

    def handle_error_no_repo(self, payload):
        print(clr.error(f'Mod repo not initialized. Use "{clr.cli("carrot init")}".'))

    def handle_error_mod_key_not_found(self, payload):
        print(clr.error('Mod key not found:'), clr.mod_key(payload.mod_key))

    def handle_error_no_mod_key_match(self, payload):
        print(f'No matches found, please verify mod key specified or use "{clr.cli("carrot search")}" to find a mod to install.')

    def handle_error_mod_not_installed(self, payload):
        print(f'No mod matching exactly the key "{clr.mod_key(payload.mod_key)}" is currently installed.')

    def handle_error_no_mods_installed(self, payload):
        print(f'No mods are installed. Use "{clr.cli("carrot install")}" to install some.')

    def handle_error_mod_file_missing(self, payload):
        print(f'Mod file for {clr.mod_name(payload.mod.name)} {clr.mod_key("[" + payload.mod.key + "]")} is {clr.error("missing")}.')


from argparse import ArgumentParser, SUPPRESS

from carrot_mc.backend import BackendService
from carrot_mc.carrot import CarrotService, InstallationManager
from carrot_mc.cli_printer import CliEventPrinter
from carrot_mc.colors import Colorizer as clr

def main():
    commands = [
        StatusCommand(),
        ListCommand(),
        InstallCommand(),
        InitCommand(),
        UpdateCommand(),
        EnableCommand(),
        DisableCommand(),
        WebGuiCommand()
    ]

    ap = ArgumentParser(prog='carrot')

    subparsers = ap.add_subparsers()

    for cmd in commands:
        cmd.register_help(subparsers)

    args = ap.parse_args()

    if hasattr(args, 'func'):
        args.func(args)
    else:
        ap.print_help()


class Command(object):
    cli_printer = CliEventPrinter()
    backend_service = BackendService()
    installation_manager = InstallationManager(backend_service, cli_printer)
    carrot_service = CarrotService(backend_service, installation_manager, cli_printer)

    def register_help(self, subparsers):
        raise NotImplementedError()

    def handle_args(self, args):
        raise NotImplementedError()


class StatusCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'status',
            help='Display a summary of current mod directory'
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        if self.carrot_service.initialized():
            carrot = self.carrot_service.get_status()

            dep_count = 0
            disabled_count = 0
            missing_file_err = []
            bad_md5_err = []

            for mod in carrot.mods:
                if mod.dependency:
                    dep_count += 1

                if mod.disabled:
                    disabled_count += 1

                if mod.file_missing:
                    missing_file_err.append(mod)
                else:
                    if mod.actual_file_md5 != mod.file.file_md5:
                        bad_md5_err.append(mod)

            print(f'Mods installed: {len(carrot.mods)}')

            if dep_count:
                print(f'of which dependencies: {dep_count}')

            if disabled_count > 0:
                print(f'Disabled mod(s): {disabled_count}')

            if len(missing_file_err) + len(bad_md5_err) == 0:
                print('All mod files seem to be in order.')
            else:
                if missing_file_err:
                    print(f'{len(missing_file_err)} mod(s) have missing files:')
                    for mod in missing_file_err:
                        print(f'\t{clr.mod_name(mod.name)}: missing file {clr.file_name(mod.file.file_name)}')

                if bad_md5_err:
                    print(f'{len(bad_md5_err)} mod(s) have possibly corrupted files:')
                    for mod in bad_md5_err:
                        actual_file_name = mod.file.file_name
                        if mod.disabled:
                            actual_file_name += '.disabled'
                        print(f'\t{clr.mod_name(mod.name)}: file {clr.file_name(actual_file_name)} has hash {clr.file_hash(mod.actual_file_md5)} instead of {clr.file_hash(mod.file.file_md5)}')

        else:
            print(clr.error('This directory does not appear to be a valid mod repo.'))


class ListCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'list',
            help='List mods installed in current directory'
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        if self.carrot_service.initialized():
            carrot = self.carrot_service.get_status()
            
            for mod in carrot.mods:
                print(f'{clr.mod_name(mod.name)} by {clr.mod_owner(mod.owner)},', end=' ')
                
                if not mod.file_missing:
                    print(f'file {clr.file_name(mod.file.file_name)}', end='')
                    
                    if mod.disabled:
                        print(clr.file_name(".disabled"), end='')
                    
                    print(' - ', end='')
                    
                    if mod.actual_file_md5 == mod.file.file_md5:
                        if not mod.disabled:
                            print(clr.ok('OK'))
                        else:
                            print(clr.warn('DISABLED'))
                    else:
                        print(clr.error('CORRUPTED'))
                
                else:
                    print(clr.error('FILE MISSING'))
                    
        else:
            print(clr.error('This directory does not appear to be a valid mod repo.'))



class InitCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'init',
            help='Initialize a mod repo in current directory.'
        )

        parser.add_argument(
            '--name',
            help='Name of this repo/modpack'
        )

        parser.add_argument(
            '--mc_version',
            help='Minecraft version',
            required=True
        )

        parser.add_argument(
            '--channel',
            help='Mod release channel',
            choices=['Alpha', 'Beta', 'Release'],
            default='Beta'
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        if self.carrot_service.initialized():
            print('ERROR! This directory is already a mod repo.')
        else:
            self.carrot_service.initialize(vars(args))
            print('Repo initialized.')


class InstallCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'install',
            help='Install a mod and its dependencies'
        )

        parser.add_argument(
            'mod_key',
            help='"Key" id of a mod *exactly* as used in e.g. CurseForge URLs. '
                 'If no mod matches exactly, a list of possible matches will be displayed.',
            nargs='+'
        )

        parser.add_argument(
            '--channel',
            help='Specify channel (Alpha/Beta/Release) to use when installing. '
                 'Overrides settings from mods.json.',
            choices=['Alpha', 'Beta', 'Release'],
            default=None
        )

        parser.add_argument(
            '--upgrade',
            help='Specify that upgrading mods to newer versions is allowed. '
                 'By default when a mod or dependency is already installed and a new version is found '
                 'during installation, it will be skipped for safety reasons.',
            nargs='?',
            const=True
        )

        parser.add_argument(
            '--downgrade',
            help='Specify that a downgrade to an older version is allowed. '
                 'This can happen e.g. if you explicitly request a Release channel but '
                 'already have a newer Beta file installed.',
            nargs='?',
            const=True
        )

        parser.set_defaults(func=self.handle_args)
    
    def handle_args(self, args):
        self.carrot_service.install(args)


class UpdateCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'update',
            help='Update a currently-installed mod (and its dependencies) or all mods to newer/older version and/or different channel.'
        )

        parser.add_argument(
            'mod_key',
            help='"Key" id of a mod to update (along with its dependencies).',
            nargs='?',
            default=None
        )

        parser.add_argument(
            '--channel',
            help='Specify target channel (Alpha/Beta/Release) to use when updating. '
                 'Overrides settings from mods.json.',
            choices=['Alpha', 'Beta', 'Release'],
            default=None
        )

        # Hidden "option" for compatibility in the installer logic
        parser.add_argument(
            '--upgrade',
            help=SUPPRESS,
            nargs='?',
            default=True
        )

        parser.add_argument(
            '--downgrade',
            help='Specify that downgrades to an older version are allowed. '
                 'This can happen e.g. if you explicitly request a Release channel but '
                 'already have newer versions of mods from the Beta channel installed.',
            nargs='?',
            const=True
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        self.carrot_service.update(args)


class EnableCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'enable',
            help='Enable a mod (by removing its ".disabled" file name suffix).'
        )

        parser.add_argument(
            'mod_key',
            help='"Key" id of a mod to enable.',
            nargs='+',
            default=None
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        self.carrot_service.enable(args)


class DisableCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'disable',
            help='Disable a mod (by appending ".disabled" to its file name).'
        )

        parser.add_argument(
            'mod_key',
            help='"Key" id of a mod to disable.',
            nargs='+',
            default=None
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        self.carrot_service.disable(args)


class WebGuiCommand(Command):
    def register_help(self, subparsers):
        parser = subparsers.add_parser(
            'web-gui',
            help='Start an interactive web GUI interface for managing mods.'
        )

        parser.add_argument(
            '--port',
            help='TCP port on which the built-in webserver will listen for connections. Default: 8877',
            default=8877
        )

        parser.add_argument(
            '--host',
            help='Host on which the built-in webserver will listen for connections. '
                 'Default: localhost/127.0.0.1 (This means it will not be accessible from other computers (RECOMMENDED)). '
                 'Change to 0.0.0.0 to listen on all available network interfaces '
                 '(REMEMBER: there is no authentication, so everyone who opens the page will be able to manage your modpack)',
            default='localhost'
        )

        parser.set_defaults(func=self.handle_args)

    def handle_args(self, args):
        import logging
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

        print(f'Starting Web GUI on http://{args.host}:{args.port} - open this address in your browser')

        from carrot_mc.web_gui.app import run_socket_app
        run_socket_app(args)


if __name__ == '__main__':
    main()


BLACK = 0
RED = 1
GREEN = 2
YELLOW = 3
BLUE = 4
MAGENTA = 5
CYAN = 6
WHITE = 7

BRIGHT = 8

RESET = '\x1b[0m'

def color_seq(color):
    if color & BRIGHT:
        return f'\x1b[{(color & 7) + 30};1m'
    else:
        return f'\x1b[{(color & 7) + 30}m'

def colorify(text, color):
    return color_seq(color) + text + RESET


class Colorizer:
    @staticmethod
    def mod_key(mod_key):
        return colorify(mod_key, WHITE+BRIGHT)
    
    @staticmethod
    def mod_name(mod_name):
        return colorify(mod_name, YELLOW+BRIGHT)
    
    @staticmethod
    def mod_owner(mod_owner):
        return colorify(mod_owner, GREEN+BRIGHT)
    
    @staticmethod
    def mod_blurb(mod_blurb):
        return colorify(mod_blurb, WHITE)
    
    @staticmethod
    def mod_category(mod_category):
        return colorify(mod_category, BLUE+BRIGHT)
    
    @staticmethod
    def file_name(file_name):
        return colorify(file_name, CYAN)
    
    @staticmethod
    def file_hash(file_hash):
        return colorify(file_hash, MAGENTA)
    
    @staticmethod
    def url(url):
        return colorify(url, BLUE)
    
    @staticmethod
    def emph(text):
        return colorify(text, BRIGHT)
    
    @staticmethod
    def ok(text):
        return colorify(text, GREEN+BRIGHT)
    
    @staticmethod
    def warn(text):
        return colorify(text, YELLOW+BRIGHT)
    
    @staticmethod
    def error(text):
        return colorify(text, RED+BRIGHT)
    
    @staticmethod
    def cli(cli):
        return colorify(cli, MAGENTA+BRIGHT)


import json

def object_to_dict_nested(obj):
    if isinstance(obj, Automappable):
        d = {}
        for k, v in vars(obj).items():
            if isinstance(v, Autoproperty):
                d[k] = object_to_dict_nested(v.value)
        return d
    elif isinstance(obj, dict):
        return obj
    elif isinstance(obj, list):
        return [object_to_dict_nested(o) for o in obj]
    else:
        return obj


class Autoproperty(object):
    def __init__(self, prop_type, default_value=None, is_list=False):
        self.prop_type = prop_type
        self.default_value = default_value
        self.value = default_value
        self.is_list = is_list


class Automappable(object):
    def to_dict(self):
        return object_to_dict_nested(self)

    @classmethod
    def from_dict(cls, obj):
        instance = cls()

        for k, v in vars(instance).items():
            if isinstance(v, Autoproperty):
                try:
                    value = obj[k]
                except:
                    value = v.default_value

                if value and issubclass(v.prop_type, Automappable):
                    if v.is_list:
                        value = [v.prop_type.from_dict(vv) for vv in value]
                    else:
                        value = v.prop_type.from_dict(value)

                setattr(instance, k, value)

        return instance

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)
        if isinstance(attr, Autoproperty):
            return attr.value
        else:
            return attr

    def __setattr__(self, name, value):
        if isinstance(value, Autoproperty):
            super().__setattr__(name, value)
            return

        try:
            attr = super().__getattribute__(name)
            if isinstance(attr, Autoproperty):
                attr.value = value
            else:
                super().__setattr__(name, value)
        except AttributeError:
            super().__setattr__(name, value)

    def __repr__(self):
        return json.dumps(self.to_dict())